from tkinter import *
from tkinter import ttk

# TODO: investigate if tkinter installs a SIGCHLD handler, which can break GDB

class PresentationLayer():
	def __init__(self, parser):
		self.top_parser = parser
		self.root = None
		self.frame = None

		self.top_parser_name = None
		self.top_parser_address = None
		self.input_chunk = None

		self.parser_env = None

		self.backtrace = None

		self.top_parser_mem_use = None

	def set_top_parser(self,parser):
		self.top_parser = parser
		self.top_parser_name.set(self.top_parser.name)
		self.top_parser_address.set(hex(self.top_parser.address))
		self.top_parser_mem_use.set(self.extract_top_parser_mem_use(self.top_parser))

	def set_input_chunk(self, input_chunk):
		self.input_chunk.set(input_chunk)

	def set_parser_env(self, parser_env):
		self.parser_env.set(parser_env)

	def set_backtrace(self, backtrace):
		self.backtrace.set(backtrace)

	# Helper for set_top_parser
	def extract_top_parser_mem_use(self, parser):
		# Join key-value pairs into a string, and put them in a list
		mem_use_list = [": ".join((hex(k), str(v))) for k,v in parser.bytes_used.items()]
		# Return the list as string joined by newlines
		return "\n".join(mem_use_list)

	# TODO: this is a bad reimplementation of hammer-parser-backtrace
	# TODO: can be moved to set_backtrace
	def parser_stack_to_backtrace_string(self, parser_stack):
		string_stack = [str(p) for p in parser_stack.p_stack[-1:-10:-1]]

		if len(parser_stack.p_stack) > 10:
			string_stack.append("...")
			string_stack.append(str(parser_stack.p_stack[0]))
		return "\n".join(string_stack)

	# let's avoid polluting the global namespace
	def init_gui(self):
		self.root = Tk()
		self.root.title("Parser Visualization")

		self.frame = ttk.Frame(self.root, padding="12 12 12 12")
		self.frame.grid(column=0, row=0, sticky='w')
		self.root.columnconfigure(0, weight=1)
		self.root.rowconfigure(0, weight=1)

		self.top_parser_name = StringVar()
		self.top_parser_name.set(self.top_parser.name)
		self.top_parser_address = StringVar()
		self.top_parser_address.set(hex(self.top_parser.address))
		self.input_chunk = StringVar()
		self.input_chunk.set(top_level_parse.get_input_chunk())
		#self.set_top_parser(top_level_parse.peek_parser())
		#self.set_input_chunk(top_level_parse.get_input_chunk())

		self.parser_env = StringVar()
		self.parser_env.set(top_level_parse.get_parser_env())

		self.backtrace = StringVar()
		self.backtrace.set(self.parser_stack_to_backtrace_string(top_level_parse.peek_parserstack()))

		self.top_parser_mem_use = StringVar()
		self.top_parser_mem_use.set(self.extract_top_parser_mem_use(self.top_parser))

		#"Current parser" widgets
		ttk.Label(self.frame, text="Current parser", anchor='w').grid(column=1, row=1)
		ttk.Label(self.frame, textvariable=self.top_parser_name, anchor='w').grid(column=1, row=2)
		ttk.Label(self.frame, textvariable=presentation_layer.top_parser_address, anchor='w').grid(column=1, row=3)

		# Display a chunk of the input
		ttk.Label(self.frame, text="Input").grid(column=2, row=1)
		ttk.Label(self.frame, textvariable=self.input_chunk).grid(column=2, row=2)

		# ParserHierarchy widget (sequence members, parsers this one contains, etc)
		# Showing parser decisions is probably possible by capturing the return value of h_do_parse
		ttk.Label(self.frame, text="Parser hierarchy").grid(column=2, row=3)
		ttk.Label(self.frame, textvariable=self.parser_env).grid(column=2, row=4)
		#ttk.Label(self.frame, text="[Placeholder, Placeholder, Placeholder]").grid(column=2, row=4)

		ttk.Label(self.frame, text="Parser backtrace").grid(column=2, row=5)
		ttk.Label(self.frame, textvariable=self.backtrace).grid(column=2, row=6)
		#ttk.Label(self.frame, text="[Placeholder, Placeholder, Placeholder]").grid(column=2, row=6)

		ttk.Label(self.frame, text="Memory use (Arena: bytes)").grid(column=3, row=1)
		ttk.Label(self.frame, textvariable=self.top_parser_mem_use).grid(column=3, row=2)

		ttk.Button(self.frame, text="Step", command=step).grid(column=1, row=4)

	def mainloop(self):
		self.root.mainloop()

def step(*args):
	gdb.execute("hammer-parse-step 1")
	presentation_layer.set_top_parser(top_level_parse.peek_parser())
	presentation_layer.set_input_chunk(top_level_parse.get_input_chunk())
	presentation_layer.set_parser_env(top_level_parse.get_parser_env())
	presentation_layer.set_backtrace(presentation_layer.parser_stack_to_backtrace_string(top_level_parse.peek_parserstack()))

if top_level_parse is None:
	print("Please import parser name script")
else:
	presentation_layer = PresentationLayer(top_level_parse.peek_parser())
	presentation_layer.init_gui()
	presentation_layer.mainloop()

