# Stops the inferior once the parse reaches (or goes past) a particular input position
# HDoParseBreakpoint checks the position parameter to determine whether to stop execution
class HammerParseStopAtInputPos(gdb.Command):
	def __init__(self):
		super(HammerParseStopAtInputPos, self).__init__ ("hammer-parse-stop-at-pos", gdb.COMMAND_OBSCURE)

	def invoke(self, arg, from_tty):
		gdb.set_convenience_variable("hammer_stop_pos", int(arg))


HammerParseStopAtInputPos()

# Implements the "step" command for parsing. One step is one invocation of h_do_parse()
class HammerParseStep(gdb.Command):
	def __init__(self):
		super(HammerParseStep, self).__init__ ("hammer-parse-step", gdb.COMMAND_OBSCURE)

	def invoke(self, arg, from_tty):
		args = gdb.string_to_argv(arg)
		if len(args) > 0:
			try:
				steps = int(args[0])
				if steps < 1:
					raise ValueError
			except ValueError:
				print("Arugment needs to be a positive integer")
				return
		else:
			steps = 1

		print("Steps: " + str(int(steps)))
		gdb.set_convenience_variable("hammer_step_counter", int(steps))
		# Counting the stops is the responsibility of HDoParseBreakpoint
		# On each breakpoint hit, something like this is executed in HDoParseBreakpoint.stop():
		# if step_counter > 0:
		#	retval = False
		#	step_counter -= 1
		# else:
		#	retval = True

		# only execute "continue" if inferior is present and running, to avoid a warning
		# don't know if there's a more reliable way to tell this
		if gdb.selected_inferior().pid > 0:
			gdb.execute("continue")

HammerParseStep()

# Analogous to "continue", but for the parse process
class HammerParseContinue(gdb.Command):
	def __init__(self):
		super(HammerParseContinue, self).__init__ ("hammer-parse-continue", gdb.COMMAND_OBSCURE)

	def invoke(self, arg, from_tty):
		gdb.execute("continue")

HammerParseContinue()

# TODO: needs reverse stepping, which probably requires some configuration beforehand
#def HammerParseStepBack(gdb.Command):
#	def __init__(self):
#		super(HammerParseContinue, self).__init__ ("hammer-parse-step-back", gdb.COMMAND_OBSCURE)
